from pycozo.ext_impl import CozoMagics


def auto_cozo_mode(lines):
    if lines:
        if lines[0].strip() == '%%py':
            del lines[0]
        elif not (lines[0].startswith('%') or lines[0].startswith('!')):
            lines.insert(0, '%%cozo')
    return lines


def load_ipython_extension(ipython):
    ipython.input_transformers_cleanup.append(auto_cozo_mode)
    ipython.register_magics(CozoMagics(ipython))



from IPython.core.magic import (Magics, magics_class, line_magic,
                                cell_magic, needs_local_scope)

from pycozo.client import Client, QueryException


@magics_class
class CozoMagics(Magics):
    def __init__(self, shell, **kwargs):
        super().__init__(shell, **kwargs)
        self.client = None
        self.params = {}

    @line_magic
    def cozo_auth(self, line: str):
        args = line.split()
        if len(args) != 1:
            raise Exception('Wrong number of arguments: one required')
        self.client.auth = args[0]
        try:
            self.client.run('?[a] := a = 1 + 1', self.params)
        except QueryException as e:
            return e
        except Exception as e:
            return e

    @line_magic
    def cozo_host(self, line: str):
        self.client = Client(host=line.strip())

    @line_magic
    def cozo_path(self, line: str):
        self.client = Client(path=line.strip())

    @cell_magic
    def cozo(self, line, cell):
        try:
            res = self.client.run(cell, self.params)
        except QueryException as e:
            return e
        except Exception as e:
            return e
        var = line.strip()
        if var:
            self.shell.user_ns[var] = res
        try:
            return res.style.applymap(colour_code_type)
        except:
            return res

    @line_magic
    @needs_local_scope
    def cozo_run_string(self, line, local_ns):
        script = eval(line, self.shell.user_ns, local_ns)
        if not isinstance(script, str):
            raise Exception('a string is required')
        try:
            return self.client.run(script, self.params)
        except QueryException as e:
            return e

    @line_magic
    def cozo_run_file(self, line):
        filename = line.strip()
        with open(filename, encoding='utf-8') as f:
            script = f.read()
        try:
            return self.client.run(script, self.params)
        except QueryException as e:
            return e

    @line_magic
    def cozo_clear(self, _line):
        self.params.clear()

    @line_magic
    @needs_local_scope
    def cozo_set(self, line, local_ns):
        var, val = line.split(maxsplit=1)
        val = eval(val, self.shell.user_ns, local_ns)
        try:
            import pandas
            if isinstance(val, pandas.DataFrame):
                self.params[var] = val.values.tolist()
            else:
                self.params[var] = val
        except ImportError:
            self.params[var] = val

    @line_magic
    @needs_local_scope
    def cozo_set_params(self, line, local_ns):
        val = eval(line, self.shell.user_ns, local_ns)
        if isinstance(val, dict):
            self.params = val
        else:
            raise Exception('params must be specified as a dict')

    @line_magic
    def cozo_params(self, _line):
        return self.params

    @line_magic
    def cozo_client(self, _line):
        return self.client


def colour_code_type(val):
    if isinstance(val, int) or isinstance(val, float):
        colour = '#307fc1'
    elif isinstance(val, str):
        colour = 'black'
    else:
        colour = '#bf5b3d'
    return f'color: {colour}'



class Client:
    def __init__(self, *, path=None, host=None, auth=None, dataframe=True):
        self.pandas = None
        if path is not None:
            from cozo_embedded import CozoDbPy
            self.embedded = CozoDbPy(path)
        elif host is not None:
            self.host = host
            self.auth = auth or ''
        else:
            raise Exception('you must specify either `path` for embedded mode, or `host` for client/server mode')

        if dataframe:
            try:
                import pandas
                self.pandas = pandas
            except ImportError as _:
                print('`pandas` feature was requested, but pandas is not installed')
                pass

    def url(self):
        return f'{self.host}/text-query'

    def headers(self):
        return {
            'x-cozo-auth': self.auth
        }

    def client_request(self, script, params=None):
        import requests

        r = requests.post(self.url(), headers=self.headers(), json={
            'script': script,
            'params': params or {}
        })
        res = r.json()
        return self.format_return(res)

    def format_return(self, res):
        if not res['ok']:
            raise QueryException(res)

        if self.pandas:
            return self.pandas.DataFrame(columns=res['headers'], data=res['rows'])
        else:
            return res

    def embedded_request(self, script, params=None):
        import json

        params_str = json.dumps(params or {}, ensure_ascii=False)
        res = self.embedded.run_query(script, params_str)
        res = json.loads(res)
        return self.format_return(res)

    def run(self, script, params=None):
        if self.embedded is None:
            return self.client_request(script, params)
        else:
            return self.embedded_request(script, params)


class QueryException(Exception):
    def __init__(self, resp):
        super().__init__()
        self.resp = resp

    def __repr__(self):
        return self.resp.get('display') or self.resp.get('message') or str(self.resp)

    def __str__(self):
        return self.resp.get('message') or str(self.resp)

    def _repr_pretty_(self, p, cycle):
        p.text(repr(self))



from setuptools import setup

setup(
    name='pycozo',
    version='0.1.5',
    packages=['pycozo'],
    url='',
    license='MIT',
    author='Ziyang Hu',
    author_email='hu.ziyang@cantab.net',
    description='Python client for the Cozo database',
    install_requires=[],
    extras_require={
        'pandas': ['pandas', 'ipython'],
        'embedded': ['cozo-embedded==0.1.4'],
        'client': ['requests']
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Framework :: IPython",
        "Programming Language :: Python",
        "Topic :: Utilities",
    ],

)
